from abc import ABC, abstractmethod


class BaseContainer(ABC):

    def __init__(self, storage, stats=None):
        self._storage = storage
        self._stats = stats

    @abstractmethod
    def get(self):
        raise NotImplementedError()

    def _report_stats_action(self, action_name, counter=1):
        if self._stats is None:
            return

        name = '%s_%s' % (self.__class__.__name__, action_name)

        if name not in self._stats:
            self._stats[name] = 0

        self._stats[name] += counter



import inspect
from abc import ABC, abstractmethod

from flying_ioc.base_container import BaseContainer


class IocFactory(ABC):
    @staticmethod
    @abstractmethod
    def get_instance(ioc_manager, name: str, frame_info: inspect.FrameInfo):
        raise NotImplementedError()


class FactoryContainer(BaseContainer):

    def get(self):
        stack = inspect.stack()
        i = 0
        while i < len(stack):
            slf = stack[i].frame.f_locals.get('self')
            i += 1
            if slf and slf.__class__.__name__ == 'IocManager':
                break

        frame_info = stack[i]
        return self._storage.get_instance(ioc_manager=self._ioc_manager, name=self._name, frame_info=frame_info)

    def __init__(self, name, cls, ioc_manager, stats=None):
        if not issubclass(cls, IocFactory):
            raise AttributeError('Argument cls should be a subclass of Factory')
        super().__init__(storage=cls, stats=stats)
        self._ioc_manager = ioc_manager
        self._name = name



from flying_ioc.class_container import ClassContainer


class SingletonContainer(ClassContainer):
    def __init__(self, cls, ioc_manager, stats=None, thread_local=False):
        super().__init__(cls=cls, ioc_manager=ioc_manager, stats=stats, thread_local=thread_local)

    def get(self):
        instance = getattr(self._instance, 'value', None)
        if instance is not None:
            self._report_stats_action('get')
            return instance

        self._instance.value = super().get()
        self._report_stats_action('store')
        return self._instance.value



from typing import List


class NotInject:
    _not_inject_classes = {}

    def __init__(self, names: List[str]):
        self._names = names

    def __call__(self, cls):
        self._not_inject_classes[cls] = self._names
        return cls

    @staticmethod
    def get_not_injected_names(cls: type):
        return NotInject._not_inject_classes.get(cls, [])



import inspect
import threading

from flying_ioc.base_container import BaseContainer
from flying_ioc import NotInject


class ClassContainer(BaseContainer):
    def __init__(self, cls, ioc_manager, stats=None, thread_local=False):
        super().__init__(storage=cls, stats=stats)
        self._ioc_manager = ioc_manager
        self._instance = threading.local() if thread_local else self

    def get(self):
        init_args = getattr(self._instance, 'init_args', None)
        if init_args is not None:
            return self.__create_class_instance(init_args)

        init_args = {}

        not_inject = NotInject.get_not_injected_names(self._storage)
        args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations = \
            self.__fill_args(self._storage, init_args, not_inject=not_inject)

        # if **kwargs is a param in the constructor
        if kwonlyargs is not None:
            self.__fill_base_classes_args(self._storage, init_args, not_inject)

        ret_val = self.__create_class_instance(init_args)
        self._instance.init_args = init_args
        return ret_val

    def __fill_args(self, cls, init_args, not_inject, raise_error=True):
        args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations = inspect.getfullargspec(cls)

        for arg in args:
            if arg == 'self':
                continue

            if arg in init_args:
                continue

            if arg in not_inject:
                continue

            arg_class = annotations.get(arg)

            if arg_class is not None:
                if arg_class.__name__ in not_inject:
                    continue
                try:
                    value = getattr(self._ioc_manager, arg_class.__name__)
                    init_args[arg] = value
                    continue
                except AttributeError:
                    pass

            try:
                value = getattr(self._ioc_manager, arg)
                init_args[arg] = value
            except AttributeError:
                if raise_error:
                    raise AttributeError("Can't get a container neither by class name %s, neither by arg name %s"
                                         % (arg_class.__name__, arg))

        return args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations

    def __fill_base_classes_args(self, cls, init_args, not_inject):
        for c in cls.__bases__:
            args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations = \
                self.__fill_args(c, init_args, not_inject=not_inject, raise_error=False)

            if kwonlyargs is not None:
                self.__fill_base_classes_args(c, init_args, not_inject=not_inject)

    def __create_class_instance(self, init_args):
        try:
            ret_val = self._storage(**init_args)
        except TypeError as e:
            print('TypeError when constructing %s: %s' % (self._storage.__name__, e.args))
            print('%s(%s)' % (self._storage.__name__,
                              ', '.join(['%s=%s' % (key, value) for key, value in init_args.items()])
                              ))
            raise e
        self._report_stats_action('create_instance')
        self._report_stats_action('get')
        return ret_val



from flying_ioc.base_container import BaseContainer


class ValueContainer(BaseContainer):
    def __init__(self, value, stats):
        super().__init__(storage=value, stats=stats)
        self._report_stats_action('store')

    def get(self):
        self._report_stats_action('get')
        return self._storage



from flying_ioc.class_container import ClassContainer
from flying_ioc.factory_container import FactoryContainer
from flying_ioc.singleton_container import SingletonContainer
from flying_ioc.value_container import ValueContainer


class IocManager:
    def __init__(self, stats=False):
        self._containers = {}
        self._stats = {} if stats else None
        self.set_value('IocManager', self)

    def __getattr__(self, name):
        if name not in self._containers:
            raise AttributeError("Name '%s' does not exist" % name)
        try:
            return self._containers[name].get()
        except AttributeError as e:
            # traceback.print_exception(e, e, e.__traceback__)
            raise TypeError(e)

    def set_value(self, name: str, value):
        self.__store_by_name(name, ValueContainer(value=value, stats=self._stats))

    def set_class(self, cls: type, name: str = None, singleton=False, thread_local=False):
        container = SingletonContainer if singleton else ClassContainer
        if name is None:
            name = cls.__name__
        self.__store_by_name(
            name,
            container(cls=cls, ioc_manager=self, stats=self._stats, thread_local=thread_local)
        )

    def set_factory(self, name: str, cls):
        self.__store_by_name(name, FactoryContainer(name=name, cls=cls, ioc_manager=self, stats=self._stats))

    def __store_by_name(self, name: str, value):
        if name in self._containers:
            raise AttributeError("Name '%s' is already defined" % name)
        self._containers[name] = value

    def print_stats(self):
        if self._stats:
            for key, value in self._stats.items():
                print('%s=%d' % (key, value))
